==> 0.0.21.tar.gz: lib/src/network.dart <==
// ignore_for_file: unnecessary_type_check

import 'dart:convert';
import 'dart:io';

// import 'package:cool';
import 'package:cookie_jar/cookie_jar.dart';
import 'package:dio/adapter.dart';
import 'package:dio/dio.dart';

import '../igapi.dart';
import 'actions.dart';
import 'device.dart';
import 'config.dart';
import 'makers.dart';
import 'quark_ig.dart';

enum HTTPBodyConsumes { FORM, JSON, FORM_DATA, TEXT, BLOB }

typedef void NetResolve(IGAPIAction? action);
typedef void NetReject(IGAPIAction? action, dynamic e);

abstract class NetBase {
  late Dio _dio;

  final QuarkDevice device;
  final String userAgent;

  NetBase(this.device, String baseUrl)
      : userAgent = IGAPIConfig.getUserAgent(device) {
    final headers = IGAPIConfig.commonHeaders;
    headers["User-Agent"] = userAgent;
    headers["Accept"] = "*/*";
    // Check localized message for debug
    // headers["Accept-Language"] = "zh-Hans-US;q=1, en-US;q=0.9";
    headers["Accept-Language"] = "en-US;q=1";
    
    _dio = new Dio(BaseOptions(
        baseUrl: baseUrl,
        connectTimeout: 10000,
        receiveTimeout: 10000,
        headers: headers,
        validateStatus: (int? a) => true));
    if (onIGHttpClientCreate != null) {
      final adapter = _dio.httpClientAdapter as DefaultHttpClientAdapter;
      adapter.onHttpClientCreate = onIGHttpClientCreate;
    }
  }

  NetResolve? netResolve;
  NetReject?  netReject;

  Future<SXEndpointModel> _fetch(String method, String? path,
      {bool?directory,
      String? query,
      Map<String, dynamic>? queryItems,
      Map<String, String>? headers,
      HTTPBodyConsumes? consumes,
      dynamic body,
      IGAPIAction? action}) async {
    path = path ?? "";

    if (directory == true) path = path + "/";
    if (query != null) {
      path = path + "?" + query;
      // 当使用quey时屏蔽queryItems
      queryItems = null;
    }

    String contentType = "";
    switch (consumes) {
      case HTTPBodyConsumes.FORM:
        contentType = "application/x-www-form-urlencoded";
        break;
      case HTTPBodyConsumes.JSON:
        contentType = "json";
        break;
      case HTTPBodyConsumes.FORM_DATA:
        // do nothing, body is formdata
        break;
      case HTTPBodyConsumes.TEXT:
        contentType = "text";
        break;
      case HTTPBodyConsumes.BLOB:
        contentType = "binary";
        break;
      default:
        break;
    }

    try {
      final res = await _dio.request(path,
        queryParameters: queryItems,
        data: body,
        options: new Options(
            method: method,
            headers: headers,
            contentType: contentType,
            extra: {"action": action}));
      res.extra = res.extra;
      
      if (netResolve != null) netResolve!(action);
      return Future.value(res.extra["ep"] ?? SXEndpointModel());
    } catch(e) {
      if (netReject != null) netReject!(action, e);
      rethrow;
    }
  }

  Future<SXEndpointModel> get(String? path,
          {bool directory = true,
          String? query,
          Map<String, dynamic>? queryItems,
          Map<String, String>? headers,
          IGAPIAction? action}) =>
      _fetch("GET", path,
          directory: directory,
          query: query,
          queryItems: queryItems,
          headers: headers,
          action: action);

  Future<SXEndpointModel> post(String? path, dynamic body,
          {bool directory = true,
          String? query,
          Map<String, dynamic>? queryItems,
          Map<String, String>? headers,
          HTTPBodyConsumes? consumes,
          IGAPIAction? action}) =>
      _fetch("POST", path,
          directory: directory,
          query: query,
          queryItems: queryItems,
          headers: headers,
          consumes: consumes,
          body: body,
          action: action);

  Future<SXEndpointModel> postForm(String path, Map<String, String> body,
          {bool directory = true,
          String? query,
          Map<String, dynamic>? queryItems,
          Map<String, String>? headers,
          IGAPIAction? action}) =>
      post(path, body,
          directory: directory,
          query: query,
          queryItems: queryItems,
          headers: headers,
          consumes: HTTPBodyConsumes.FORM,
          action: action);

  Future<SXEndpointModel> postJson(String path, Map<String, dynamic> body,
          {bool directory = true,
          String? query,
          Map<String, dynamic>? queryItems,
          Map<String, String>? headers,
          IGAPIAction? action}) =>
      post(path, body,
          directory: directory,
          query: query,
          queryItems: queryItems,
          headers: headers,
          consumes: HTTPBodyConsumes.JSON,
          action: action);

  Future<SXEndpointModel> postSignedJson(String? path, Map<String, dynamic> body,
          {bool directory = true,
          String? query,
          Map<String, dynamic>? queryItems,
          Map<String, String>? headers,
          IGAPIAction? action}) =>
      post(path, IGAPIConfig.signatureBody(body),
          directory: directory,
          query: query,
          queryItems: queryItems,
          headers: headers,
          consumes: HTTPBodyConsumes.FORM,
          action: action);

  Future<SXEndpointModel> postFormData(String path, Map<String, dynamic> body,
          {bool directory = true,
          String? query,
          Map<String, dynamic>? queryItems,
          Map<String, String>? headers,
          IGAPIAction? action}) =>
      post(path, FormData.fromMap(body),
          directory: directory,
          query: query,
          queryItems: queryItems,
          headers: headers,
          consumes: HTTPBodyConsumes.FORM_DATA,
          action: action);

  Future<SXEndpointModel> postText(String path, String text,
          {bool directory = true,
          String? query,
          Map<String, dynamic>? queryItems,
          Map<String, String>? headers,
          IGAPIAction? action}) =>
      post(path, text,
          directory: directory,
          query: query,
          queryItems: queryItems,
          headers: headers,
          consumes: HTTPBodyConsumes.TEXT,
          action: action);

  Future<SXEndpointModel> delete(String path,
          {bool directory = true,
          String? query,
          Map<String, dynamic>? queryItems,
          Map<String, String>? headers,
          IGAPIAction? action}) =>
      _fetch("DELETE", path,
          directory: directory,
          query: query,
          queryItems: queryItems,
          headers: headers,
          action: action);
}

class _IGCookieManager extends Interceptor {
  /// Cookie manager for http requests。Learn more details about
  /// CookieJar please refer to [cookie_jar](https://github.com/flutterchina/cookie_jar)
  final CookieJar cookieJar;

  static const invalidCookieValue = "_invalid_";

  /// Dart SDK will cause an exception When response cookie's value is empty,
  /// eg. 'Set-Cookie: session=; Path=/; Expires=Thu, 01 Jan 1970 00:00:00 GMT'
  ///
  /// This is a bug of Dart SDK: https://github.com/dart-lang/sdk/issues/35804
  /// So, we should normalize the cookie value before this bug is fixed.
  bool needNormalize = false;

  _IGCookieManager(this.cookieJar) {
    // Set `needNormalize` value by Duck test
    try {
      Cookie.fromSetCookieValue("k=;");
    } catch (e) {
      needNormalize = true;
    }
  }

  @override
  void onRequest(RequestOptions options, RequestInterceptorHandler handler) async {
    final cookies = await cookieJar.loadForRequest(options.uri)..removeWhere((cookie) {
      if (cookie.expires != null) {
        return cookie.expires!.isBefore(DateTime.now());
      }
      return false;
    });
    final cookie = getCookies(cookies);
    if (cookie.isNotEmpty) options.headers[HttpHeaders.cookieHeader] = cookie;
  }

  @override
  void onResponse(Response response, ResponseInterceptorHandler handler) {
    _saveCookies(response);
  }

  @override
  void onError(DioError err, ErrorInterceptorHandler handler) {
    _saveCookies(err.response);
  }

  _saveCookies(Response? response) {
    if (response != null) {
      List<String> cookies = response.headers[HttpHeaders.setCookieHeader] ?? [];
      if (needNormalize) {
        var _cookies = normalizeCookies(cookies);
        cookies
          ..clear()
          ..addAll(_cookies);
      }
      cookieJar.saveFromResponse(
        response.realUri,
        cookies.map((str) {
          try {
            return Cookie.fromSetCookieValue(str);
          } catch (e) {
            return Cookie.fromSetCookieValue(
                "__placeholder__=$invalidCookieValue; path=/; expires=Thu, 01 Jan 1970 00:00:00 GMT");
          }
        }).toList(),
      );
    }
  }

  static String getCookies(List<Cookie> cookies) {
    return cookies.map((cookie) => "${cookie.name}=${cookie.value}").join('; ');
  }

  static List<String> normalizeCookies(List<String>? cookies) {
    const String expires = " Expires=Thu, 01 Jan 1970 00:00:00 GMT";
    return cookies?.map((cookie) {
      var _cookie = cookie.split(";");
      var kv = _cookie.first.split("=");
      if (kv[1].isEmpty) {
        kv[1] = invalidCookieValue;
        _cookie[0] = kv.join('=');
        if (_cookie.length > 1) {
          int i = 1;
          while (i < _cookie.length) {
            if (_cookie[i].trim().toLowerCase().startsWith("expires")) {
              _cookie[i] = expires;
              break;
            }
            ++i;
          }
          if (i == _cookie.length) {
            _cookie.add(expires);
          }
        }
      }
      return _cookie.join(";");
    }).toList() ?? [];
  }
}

// 执行流: 请求拦截器 >> 请求转换器 >> 发起请求 >> 响应转换器 >> 响应拦截器 >> 最终结果
class _IGRemoteHandler extends Interceptor {
  final IGAccount _account;

  _IGRemoteHandler(this._account);

  Future<void> _handleBadResponse(Response response) async {
    final json = response.data;
    if (json is! Map<String, dynamic>) {
      throw IGException(IGExceptionType.UnexpectedFormat, "invalid json",
          {"statusCode": response.statusCode});
    }
    json["statusCode"] = response.statusCode;

    String status = json["status"];
    if (status != "fail") {
      throw IGException(
          IGExceptionType.AbnormalErrorStatus, "status is not fail", json);
    }

    String? message = json["message"];
    String? title = json["error_title"];
    String type = json["error_type"];

    IGExceptionType reason;
    if (json["two_factor_required"] == true) {
      reason = IGExceptionType.TwoFactorRequired;
      message = "tfa required";
      _onAuthTwoFactor(json);
    } else if (json["invalid_credentials"] == true) {
      reason = IGExceptionType.InvalidCredentials;
      message = title;
    } else if (type == "sms_code_validation_code_invalid") {
      reason = IGExceptionType.InvalidSmsCode;
      message = message ?? "invalid sms code";
    } else if (type == "invalid_nonce") {
      reason = IGExceptionType.InvalidNonce;
      message = message ?? "invalid nonce";
    } else if (type == "checkpoint_challenge_required" ||
        message == "challenge_required") {
      reason = IGExceptionType.ChallengeRequired;
      message = "challenge required";
      _onAuthChallenge(json);
    } else if (message == "login_required") {
      reason = IGExceptionType.LoginRequired;
      message = title ?? "login required";
      await _onAuthExpired();
    } else if (message == "feedback_required") {
      reason = IGExceptionType.FeedbackRequired;
      message = json["feedback_message"] ?? message;
    } else if (message != null) {
      reason = IGExceptionType.APIRejected;
      message = message;
    } else {
      reason = IGExceptionType.Uncaught;
      message = "unknwon message";
      assert(false);
    }

    throw IGException(reason, message ?? "", json);
  }

  void _onAuthTwoFactor(Map<String, dynamic> json) {
    _account.state.twoFactorLock = MM.create(IGTwoFactorLock, json);
  }

  void _onAuthChallenge(Map<String, dynamic> json) {
    _account.state.challengeLock = MM.create(IGChallengeLock, json);
  }

  Future<void> _onAuthExpired() async {
    await _account.onAuthExpired();
    return Future.value();
  }

  Future<void> _refreshAuthStatus(IGAPIAction action, SXEndpointModel ep) async {
    switch (action) {
      case IGAPIAction.AuthLogin:
      case IGAPIAction.AuthTFLogin:
        await _account.onLoggedIn(SXTuple1<IGLoggedInUser>(ep).item0);
        break;
      case IGAPIAction.AuthFBLogin:
      case IGAPIAction.AuthChallengeVerify:
        await _account.onLoggedIn(SXTuple1<IGLoggedInUser>(ep).item0, true);
        break;
      case IGAPIAction.AuthTFSmsSend:
        _account.state.twoFactorLock = SXTuple1<IGTwoFactorLock>(ep).item0;
        break;
      default:
        break;
    }
    return Future.value();
  }

  Future<void> _transformEndpointModel(Response response) async {
    response.extra["ep"] = SXEndpointModel(); // placeholder
    IGAPIAction? action = response.extra["action"];
    dynamic data = response.data;
    if (data is String) {
      final ms = json.decode(data);
      if (ms != null) data = ms;
    }
    if (action is IGAPIAction && data is Map<String, dynamic>) {
      SXEndpointModel? ep = MM.createEP(action, data);
      if (ep is SXEndpointModel) {
        await _refreshAuthStatus(action, ep);
        response.extra["ep"] = ep;
      }
    }
    return Future.value();
  }

  @override
  void onResponse(Response response, ResponseInterceptorHandler handler) async {
  if (response.statusCode == 200) {
        await _transformEndpointModel(response);
        return super.onResponse(response, handler);
      }
      if ((response.statusCode ?? 0) >= 500) {
        throw IGException(IGExceptionType.ServerInternal,
            "server internal fatal.", {"statusCode": response.statusCode});
      }

      if ((response.statusCode ?? 0) >= 400 && (response.statusCode ?? 0) < 500) {
        try {
          await _handleBadResponse(response);
        }catch(e) {
          throw e;
        }
      }
      return super.onResponse(response, handler);
  }

  @override
  void onError(DioError err, ErrorInterceptorHandler handler) {
    if (err.type != DioErrorType.connectTimeout || err.error is! IGException)
      return super.onError(err, handler);
    IGException error = err.error;
    print("IG_ERROR:$error");
    return super.onError(err, handler);
  }
}

class APINet extends NetBase {
  APINet(IGAccount account) : super(account.device, IGAPIConfig.API_V1) {
    _dio.interceptors.add(_IGCookieManager(account.cookieJar));
    _dio.interceptors.add(_IGRemoteHandler(account));
  }
}

==> 0.0.23.tar.gz: lib/src/network.dart <==
import 'dart:convert';
import 'dart:ffi';
import 'dart:io';

// import 'package:cool';
import 'package:cookie_jar/cookie_jar.dart';
import 'package:dio/adapter.dart';
import 'package:dio/dio.dart';

import '../igapi.dart';
import 'actions.dart';
import 'device.dart';
import 'config.dart';
import 'makers.dart';
import 'quark_ig.dart';

enum HTTPBodyConsumes { FORM, JSON, FORM_DATA, TEXT, BLOB }

typedef void NetResolve(IGAPIAction? action);
typedef void NetReject(IGAPIAction? action, dynamic e);

abstract class NetBase {
  Dio? _dio;

  final QuarkDevice device;
  final String userAgent;

  NetBase(this.device, String baseUrl)
      : userAgent = IGAPIConfig.getUserAgent(device) {
    final headers = IGAPIConfig.commonHeaders;
    headers["User-Agent"] = userAgent;
    headers["Accept"] = "*/*";
    // Check localized message for debug
    // headers["Accept-Language"] = "zh-Hans-US;q=1, en-US;q=0.9";
    headers["Accept-Language"] = "en-US;q=1";

    final Dio dio = Dio();
    final adapter = dio.httpClientAdapter as DefaultHttpClientAdapter;
    adapter.onHttpClientCreate = (client) {
      client.findProxy = (uri) {
        return "PROXY 192.168.124.2:8080";
        // return "DIRECT";
        // return "PROXY localhost:8080";
      };
      client.badCertificateCallback = (cert, host, port) {
        return true;
      };
      return client;
    };
    dio.options.baseUrl = baseUrl;
    dio.options.connectTimeout = 10000;
    dio.options.receiveTimeout = 10000;
    dio.options.headers = headers;
    // dio.options.contentType = "application/json";
    dio.options.responseType = ResponseType.json;
    dio.options.validateStatus = (int? status) {
      return true;
    };
    _dio = dio;
    _dio!.httpClientAdapter = adapter;

    // _dio = new Dio(BaseOptions(
    //     baseUrl: baseUrl,
    //     connectTimeout: 10000,
    //     receiveTimeout: 10000,
    //     headers: headers,
    //     validateStatus: (int? a) => true));
    // if (onIGHttpClientCreate != null) {
    //   (_dio?.httpClientAdapter as DefaultHttpClientAdapter).onHttpClientCreate =
    //       (client) {
    //     client.findProxy = (url) {
    //       // return "PROXY 192.168.124.2:8080";
    //       return "DIRECT";
    //     };
    //     client.badCertificateCallback = (cert, host, port){
    //       return true;
    //     };
    //     return client;
    //   };
    //   // final adapter = _dio.httpClientAdapter as DefaultHttpClientAdapter;
    //   // adapter.onHttpClientCreate = onIGHttpClientCreate;
    // }
  }

  NetResolve? netResolve;
  NetReject? netReject;

  Future<SXEndpointModel> _fetch(String method, String? path,
      {bool? directory,
      String? query,
      Map<String, dynamic>? queryItems,
      Map<String, String>? headers,
      HTTPBodyConsumes? consumes,
      dynamic body,
      IGAPIAction? action}) async {
    path = path ?? "";

    if (directory == true) path = path + "/";
    if (query != null) {
      path = path + "?" + query;
      // 当使用quey时屏蔽queryItems
      queryItems = null;
    }

    String contentType = "";
    switch (consumes) {
      case HTTPBodyConsumes.FORM:
        contentType = "application/x-www-form-urlencoded";
        break;
      case HTTPBodyConsumes.JSON:
        contentType = "json";
        break;
      case HTTPBodyConsumes.FORM_DATA:
        // do nothing, body is formdata
        break;
      case HTTPBodyConsumes.TEXT:
        contentType = "text";
        break;
      case HTTPBodyConsumes.BLOB:
        contentType = "binary";
        break;
      default:
        break;
    }

    try {
      final res = await _dio!.request(path,
      queryParameters: queryItems,
      data: body,
      options: new Options(
        method: method,
        headers: headers,
        contentType: contentType,
        extra: {"action": action}
      ));
      res.extra = res.extra;

      if (netResolve != null) netResolve!(action);
      return Future.value(res.extra["ep"] ?? SXEndpointModel());
    } catch (e) {
      if (netReject != null) netReject!(action, e);
      rethrow;
    }
  }

  Future<SXEndpointModel> get(String? path,
          {bool directory = true,
          String? query,
          Map<String, dynamic>? queryItems,
          Map<String, String>? headers,
          IGAPIAction? action}) =>
      _fetch("GET", path,
          directory: directory,
          query: query,
          queryItems: queryItems,
          headers: headers,
          action: action);

  Future<SXEndpointModel> post(String? path, dynamic body,
          {bool directory = true,
          String? query,
          Map<String, dynamic>? queryItems,
          Map<String, String>? headers,
          HTTPBodyConsumes? consumes,
          IGAPIAction? action}) =>
      _fetch("POST", path,
          directory: directory,
          query: query,
          queryItems: queryItems,
          headers: headers,
          consumes: consumes,
          body: body,
          action: action);

  Future<SXEndpointModel> postForm(String path, Map<String, String> body,
          {bool directory = true,
          String? query,
          Map<String, dynamic>? queryItems,
          Map<String, String>? headers,
          IGAPIAction? action}) =>
      post(path, body,
          directory: directory,
          query: query,
          queryItems: queryItems,
          headers: headers,
          consumes: HTTPBodyConsumes.FORM,
          action: action);

  Future<SXEndpointModel> postJson(String path, Map<String, dynamic> body,
          {bool directory = true,
          String? query,
          Map<String, dynamic>? queryItems,
          Map<String, String>? headers,
          IGAPIAction? action}) =>
      post(path, body,
          directory: directory,
          query: query,
          queryItems: queryItems,
          headers: headers,
          consumes: HTTPBodyConsumes.JSON,
          action: action);

  Future<SXEndpointModel> postSignedJson(
          String? path, Map<String, dynamic> body,
          {bool directory = true,
          String? query,
          Map<String, dynamic>? queryItems,
          Map<String, String>? headers,
          IGAPIAction? action}) =>
      post(path, IGAPIConfig.signatureBody(body),
          directory: directory,
          query: query,
          queryItems: queryItems,
          headers: headers,
          consumes: HTTPBodyConsumes.FORM,
          action: action);

  Future<SXEndpointModel> postFormData(String path, Map<String, dynamic> body,
          {bool directory = true,
          String? query,
          Map<String, dynamic>? queryItems,
          Map<String, String>? headers,
          IGAPIAction? action}) =>
      post(path, FormData.fromMap(body),
          directory: directory,
          query: query,
          queryItems: queryItems,
          headers: headers,
          consumes: HTTPBodyConsumes.FORM_DATA,
          action: action);

  Future<SXEndpointModel> postText(String path, String text,
          {bool directory = true,
          String? query,
          Map<String, dynamic>? queryItems,
          Map<String, String>? headers,
          IGAPIAction? action}) =>
      post(path, text,
          directory: directory,
          query: query,
          queryItems: queryItems,
          headers: headers,
          consumes: HTTPBodyConsumes.TEXT,
          action: action);

  Future<SXEndpointModel> delete(String path,
          {bool directory = true,
          String? query,
          Map<String, dynamic>? queryItems,
          Map<String, String>? headers,
          IGAPIAction? action}) =>
      _fetch("DELETE", path,
          directory: directory,
          query: query,
          queryItems: queryItems,
          headers: headers,
          action: action);
}

class _IGCookieManager extends Interceptor {
  /// Cookie manager for http requests。Learn more details about
  /// CookieJar please refer to [cookie_jar](https://github.com/flutterchina/cookie_jar)
  final CookieJar cookieJar;

  static const invalidCookieValue = "_invalid_";

  /// Dart SDK will cause an exception When response cookie's value is empty,
  /// eg. 'Set-Cookie: session=; Path=/; Expires=Thu, 01 Jan 1970 00:00:00 GMT'
  ///
  /// This is a bug of Dart SDK: https://github.com/dart-lang/sdk/issues/35804
  /// So, we should normalize the cookie value before this bug is fixed.
  bool needNormalize = false;

  _IGCookieManager(this.cookieJar) {
    // Set `needNormalize` value by Duck test
    try {
      Cookie.fromSetCookieValue("k=;");
    } catch (e) {
      needNormalize = true;
    }
  }

  @override
  void onRequest(
      RequestOptions options, RequestInterceptorHandler handler) async {
    final cookies = await cookieJar.loadForRequest(options.uri)
      ..removeWhere((cookie) {
        if (cookie.expires != null) {
          return cookie.expires!.isBefore(DateTime.now());
        }
        return false;
      });
    final cookie = getCookies(cookies);
    if (cookie.isNotEmpty) options.headers[HttpHeaders.cookieHeader] = cookie;
    super.onRequest(options, handler);
  }

  @override
  void onResponse(Response response, ResponseInterceptorHandler handler) {
    _saveCookies(response);
    super.onResponse(response, handler);
  }

  @override
  void onError(DioError err, ErrorInterceptorHandler handler) {
    _saveCookies(err.response);
    super.onError(err, handler);
  }

  _saveCookies(Response? response) {
    if (response != null) {
      List<String> cookies =
          response.headers[HttpHeaders.setCookieHeader] ?? [];
      if (needNormalize) {
        var _cookies = normalizeCookies(cookies);
        cookies
          ..clear()
          ..addAll(_cookies);
      }
      cookieJar.saveFromResponse(
        response.realUri,
        cookies.map((str) {
          try {
            return Cookie.fromSetCookieValue(str);
          } catch (e) {
            return Cookie.fromSetCookieValue(
                "__placeholder__=$invalidCookieValue; path=/; expires=Thu, 01 Jan 1970 00:00:00 GMT");
          }
        }).toList(),
      );
    }
  }

  static String getCookies(List<Cookie> cookies) {
    return cookies.map((cookie) => "${cookie.name}=${cookie.value}").join('; ');
  }

  static List<String> normalizeCookies(List<String>? cookies) {
    const String expires = " Expires=Thu, 01 Jan 1970 00:00:00 GMT";
    return cookies?.map((cookie) {
          var _cookie = cookie.split(";");
          var kv = _cookie.first.split("=");
          if (kv[1].isEmpty) {
            kv[1] = invalidCookieValue;
            _cookie[0] = kv.join('=');
            if (_cookie.length > 1) {
              int i = 1;
              while (i < _cookie.length) {
                if (_cookie[i].trim().toLowerCase().startsWith("expires")) {
                  _cookie[i] = expires;
                  break;
                }
                ++i;
              }
              if (i == _cookie.length) {
                _cookie.add(expires);
              }
            }
          }
          return _cookie.join(";");
        }).toList() ??
        [];
  }
}

// 执行流: 请求拦截器 >> 请求转换器 >> 发起请求 >> 响应转换器 >> 响应拦截器 >> 最终结果
class _IGRemoteHandler extends Interceptor {
  final IGAccount _account;

  _IGRemoteHandler(this._account);

  Future<void> _handleBadResponse(Response response) async {
    final json = response.data;
    if (json is! Map<String, dynamic>) {
      throw IGException(IGExceptionType.UnexpectedFormat, "invalid json",
          {"statusCode": response.statusCode});
    }
    json["statusCode"] = response.statusCode;

    String status = json["status"];
    if (status != "fail") {
      throw IGException(
          IGExceptionType.AbnormalErrorStatus, "status is not fail", json);
    }

    String? message = json["message"];
    String? title = json["error_title"];
    String type = json["error_type"];

    IGExceptionType reason;
    if (json["two_factor_required"] == true) {
      reason = IGExceptionType.TwoFactorRequired;
      message = "tfa required";
      _onAuthTwoFactor(json);
    } else if (json["invalid_credentials"] == true) {
      reason = IGExceptionType.InvalidCredentials;
      message = title;
    } else if (type == "sms_code_validation_code_invalid") {
      reason = IGExceptionType.InvalidSmsCode;
      message = message ?? "invalid sms code";
    } else if (type == "invalid_nonce") {
      reason = IGExceptionType.InvalidNonce;
      message = message ?? "invalid nonce";
    } else if (type == "checkpoint_challenge_required" ||
        message == "challenge_required") {
      reason = IGExceptionType.ChallengeRequired;
      message = "challenge required";
      _onAuthChallenge(json);
    } else if (message == "login_required") {
      reason = IGExceptionType.LoginRequired;
      message = title ?? "login required";
      await _onAuthExpired();
    } else if (message == "feedback_required") {
      reason = IGExceptionType.FeedbackRequired;
      message = json["feedback_message"] ?? message;
    } else if (message != null) {
      reason = IGExceptionType.APIRejected;
      message = message;
    } else {
      reason = IGExceptionType.Uncaught;
      message = "unknwon message";
      assert(false);
    }

    throw IGException(reason, message ?? "", json);
  }

  void _onAuthTwoFactor(Map<String, dynamic> json) {
    _account.state.twoFactorLock = MM.create(IGTwoFactorLock, json);
  }

  void _onAuthChallenge(Map<String, dynamic> json) {
    _account.state.challengeLock = MM.create(IGChallengeLock, json);
  }

  Future<void> _onAuthExpired() async {
    await _account.onAuthExpired();
    return Future.value();
  }

  Future<void> _refreshAuthStatus(
      IGAPIAction action, SXEndpointModel ep) async {
    switch (action) {
      case IGAPIAction.AuthLogin:
      case IGAPIAction.AuthTFLogin:
        await _account.onLoggedIn(SXTuple1<IGLoggedInUser>(ep).item0);
        break;
      case IGAPIAction.AuthFBLogin:
      case IGAPIAction.AuthChallengeVerify:
        await _account.onLoggedIn(SXTuple1<IGLoggedInUser>(ep).item0, true);
        break;
      case IGAPIAction.AuthTFSmsSend:
        _account.state.twoFactorLock = SXTuple1<IGTwoFactorLock>(ep).item0;
        break;
      default:
        break;
    }
    return Future.value();
  }

  Future<void> _transformEndpointModel(Response response) async {
    response.extra["ep"] = SXEndpointModel(); // placeholder
    IGAPIAction? action = response.requestOptions.extra["action"];
    dynamic data = response.data;
    if (data is String) {
      final ms = json.decode(data);
      if (ms != null) data = ms;
    }
    if (action is IGAPIAction && data is Map<String, dynamic>) {
      SXEndpointModel? ep = MM.createEP(action, data);
      if (ep is SXEndpointModel) {
        await _refreshAuthStatus(action, ep);
        response.extra["ep"] = ep;
      }
    }
    return Future.value();
  }

  @override
  void onResponse(Response response, ResponseInterceptorHandler handler) async {
    if (response.statusCode == 200) {
      await _transformEndpointModel(response);
      return super.onResponse(response, handler);
    }
    if ((response.statusCode ?? 0) >= 500) {
      throw IGException(IGExceptionType.ServerInternal,
          "server internal fatal.", {"statusCode": response.statusCode});
    }

    if ((response.statusCode ?? 0) >= 400 && (response.statusCode ?? 0) < 500) {
      try {
        await _handleBadResponse(response);
      } catch (e) {
        throw e;
      }
    }
    return super.onResponse(response, handler);
  }

  @override
  void onError(DioError err, ErrorInterceptorHandler handler) {
    if (err.type != DioErrorType.connectTimeout || err.error is! IGException)
      return super.onError(err, handler);
    IGException error = err.error;
    print("IG_ERROR:$error");
    return super.onError(err, handler);
  }
}

class APINet extends NetBase {
  APINet(IGAccount account) : super(account.device, IGAPIConfig.API_V1) {
    _dio?.interceptors.add(_IGCookieManager(account.cookieJar));
    _dio?.interceptors.add(_IGRemoteHandler(account));
  }
}
